from __future__ import annotations

import os

import pandas as pd
from packaging import version

DEVELOPMENT = False
if DEVELOPMENT:
    # DEVELOPMENT mode is used to load and create instrument specs from a CSV file.
    # This is loaded by default and slower to parse than directly creating a dict
    # So when packaging output the INSTRUMENT_SPEC dict and paste into the non-development
    # section.

    def _append_kwargs_name(df):
        """combine the columns leg and kwargs to produce library consistent kwargs for dicts"""
        prefix = df["leg"]
        prefix = prefix.where(prefix == "leg2", "")
        prefix = prefix.replace("leg2", "leg2_")
        df["kwargs_name"] = prefix + df["kwarg"]
        return df.set_index("kwargs_name")

    def _parse_bool(df):
        """parse data input as bools to return True and False dtypes."""

        def _map_true_false(v):
            try:
                if v.upper() == "TRUE":
                    return True
                elif v.upper() == "FALSE":
                    return False
            except AttributeError:
                return None
            else:
                return None

        if version.parse(pd.__version__) >= version.parse("2.1.0"):
            # applymap issues a deprecation warning with version <2.1.0
            # TODO (low): clean this up when setting a minimum pandas version at 2.1.0
            df[df["dtype"] == "bool"] = df[df["dtype"] == "bool"].map(_map_true_false)
        else:
            df[df["dtype"] == "bool"] = df[df["dtype"] == "bool"].applymap(_map_true_false)
        return df

    path = "data/__instrument_spec.csv"
    abspath = os.path.dirname(os.path.abspath(__file__))
    target = os.path.join(abspath, path)
    df = pd.read_csv(target)
    df = _append_kwargs_name(df)
    df = _parse_bool(df)
    df_legs = df[~(df["leg"] == "meta")]

    DTYPE_MAP = {
        "str": str,
        "float": float,
        "bool": bool,
        "int": int,
    }

    def _map_dtype(v):
        try:
            return DTYPE_MAP[v]
        except KeyError:
            return v

    def _map_str_int(v):
        try:
            return int(v)
        except ValueError:
            return v

    def _get_kwargs(spec):
        """From the legs DataFrame extract the relevant column and ensure dtypes are suitable."""
        # get values that are not null
        s = df_legs[spec]
        s = s[pd.notna(s)]
        # assign the correct dtypes for the values
        dfs = s.to_frame().transpose()
        dtypes = df.loc[s.index, "dtype"]
        dtypes = dtypes.map(_map_dtype)
        dfs = dfs.astype(dtype=dtypes.to_dict(), errors="raise")
        # rotate and return values in a dict
        s = dfs.transpose()[spec]
        d = s.to_dict()

        # roll dtype is str or int causes object issues
        if "roll" in d:
            d["roll"] = _map_str_int(d["roll"])
        if "leg2_roll" in d:
            d["leg2_roll"] = _map_str_int(d["leg2_roll"])
        return d

    INSTRUMENT_SPECS = {k: _get_kwargs(k) for k in df.columns[4:]}

    # add some aliases

    INSTRUMENT_SPECS = {
        **INSTRUMENT_SPECS,
        "sofr": INSTRUMENT_SPECS["usd_irs"],
        "ust": INSTRUMENT_SPECS["us_gb"],
        "ustb": INSTRUMENT_SPECS["us_gbb"],
        "ukt": INSTRUMENT_SPECS["uk_gb"],
        "gilt": INSTRUMENT_SPECS["uk_gb"],
        "cadgb": INSTRUMENT_SPECS["ca_gb"],
        "sgb": INSTRUMENT_SPECS["se_gb"],
        "sgbb": INSTRUMENT_SPECS["se_gbb"],
        "ukti": INSTRUMENT_SPECS["gbp_gbi"],
        "sofr3mf": INSTRUMENT_SPECS["usd_stir"],
        "sofr1mf": INSTRUMENT_SPECS["usd_stir1"],
        "sonia3mf": INSTRUMENT_SPECS["gbp_stir"],
        "estr1mf": INSTRUMENT_SPECS["eur_stir1"],
        "estr3mf": INSTRUMENT_SPECS["eur_stir"],
        "euribor3mf": INSTRUMENT_SPECS["eur_stir3"],
    }

else:
    INSTRUMENT_SPECS = {
        "test": {
            "frequency": "m",
            "stub": "longfront",
            "eom": False,
            "modifier": "p",
            "calendar": "nyc,tgt,ldn",
            "payment_lag": 4,
            "currency": "tes",
            "convention": "test",
            "leg2_frequency": "m",
            "leg2_stub": "longback",
            "leg2_roll": 1,
            "leg2_eom": False,
            "leg2_modifier": "mp",
            "leg2_calendar": "nyc,tgt,ldn",
            "leg2_payment_lag": 3,
            "leg2_convention": "test2",
        },
        "usd_irs": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "nyc",
            "payment_lag": 2,
            "currency": "usd",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "usd_irs_lt_2y": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": True,
            "modifier": "mf",
            "calendar": "nyc",
            "payment_lag": 2,
            "currency": "usd",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "gbp_irs": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": True,
            "modifier": "mf",
            "calendar": "ldn",
            "payment_lag": 0,
            "currency": "gbp",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "eur_irs": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 1,
            "currency": "eur",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "sek_irs": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "stk",
            "payment_lag": 1,
            "currency": "sek",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "nok_irs": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "osl",
            "payment_lag": 2,
            "currency": "nok",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "chf_irs": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "zur",
            "payment_lag": 2,
            "currency": "chf",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "cad_irs": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "tro",
            "payment_lag": 1,
            "currency": "cad",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "cad_irs_le_1y": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": True,
            "modifier": "mf",
            "calendar": "tro",
            "payment_lag": 1,
            "currency": "cad",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "jpy_irs": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": True,
            "modifier": "mf",
            "calendar": "tyo",
            "payment_lag": 2,
            "currency": "jpy",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "aud_irs6": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": True,
            "modifier": "mf",
            "calendar": "syd",
            "payment_lag": 0,
            "currency": "aud",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "aud_irs3": {
            "frequency": "q",
            "stub": "shortfront",
            "eom": True,
            "modifier": "mf",
            "calendar": "syd",
            "payment_lag": 0,
            "currency": "aud",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "aud_irs": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": True,
            "modifier": "mf",
            "calendar": "syd",
            "payment_lag": 2,
            "currency": "aud",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "eur_irs6": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "30e360",
            "leg2_frequency": "s",
            "leg2_convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "eur_irs3": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "30e360",
            "leg2_frequency": "q",
            "leg2_convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "eur_irs1": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "30e360",
            "leg2_frequency": "m",
            "leg2_convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "sek_irs3": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "stk",
            "payment_lag": 0,
            "currency": "sek",
            "convention": "30e360",
            "leg2_frequency": "q",
            "leg2_convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "nok_irs3": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "osl",
            "payment_lag": 0,
            "currency": "nok",
            "convention": "30e360",
            "leg2_frequency": "q",
            "leg2_convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "nok_irs6": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "osl",
            "payment_lag": 0,
            "currency": "nok",
            "convention": "30e360",
            "leg2_frequency": "s",
            "leg2_convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "eurusd_xcs": {
            "frequency": "q",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt,nyc",
            "payment_lag": 2,
            "currency": "eur",
            "convention": "act360",
            "leg2_currency": "usd",
            "spread_compound_method": "none_simple",
            "fixing_method": "rfr_payment_delay",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "payment_lag_exchange": 0,
            "fixed": False,
            "leg2_fixed": False,
            "leg2_mtm": True,
        },
        "gbpusd_xcs": {
            "frequency": "q",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "ldn,nyc",
            "payment_lag": 2,
            "currency": "gbp",
            "convention": "act365f",
            "leg2_currency": "usd",
            "leg2_convention": "act360",
            "spread_compound_method": "none_simple",
            "fixing_method": "rfr_payment_delay",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "payment_lag_exchange": 0,
            "fixed": False,
            "leg2_fixed": False,
            "leg2_mtm": True,
        },
        "eurgbp_xcs": {
            "frequency": "q",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt,ldn",
            "payment_lag": 2,
            "currency": "eur",
            "convention": "act360",
            "leg2_currency": "gbp",
            "leg2_convention": "act365f",
            "spread_compound_method": "none_simple",
            "fixing_method": "rfr_payment_delay",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "payment_lag_exchange": 0,
            "fixed": False,
            "leg2_fixed": False,
            "leg2_mtm": True,
        },
        "gbpeur_xcs": {
            "frequency": "q",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt,ldn",
            "payment_lag": 2,
            "currency": "gbp",
            "convention": "act365f",
            "leg2_currency": "eur",
            "leg2_convention": "act360",
            "spread_compound_method": "none_simple",
            "fixing_method": "rfr_payment_delay",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "payment_lag_exchange": 0,
            "fixed": False,
            "leg2_fixed": False,
            "leg2_mtm": True,
        },
        "jpyusd_xcs": {
            "frequency": "q",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "nyc,tyo",
            "payment_lag": 2,
            "currency": "jpy",
            "convention": "act365f",
            "leg2_currency": "usd",
            "leg2_convention": "act360",
            "spread_compound_method": "none_simple",
            "fixing_method": "rfr_payment_delay",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "payment_lag_exchange": 0,
            "fixed": False,
            "leg2_fixed": False,
            "leg2_mtm": True,
        },
        "eur_zcis": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "1+",
            "leg2_index_method": "monthly",
            "leg2_index_lag": 3,
        },
        "gbp_zcis": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "ldn",
            "payment_lag": 0,
            "currency": "gbp",
            "convention": "1+",
            "leg2_index_method": "monthly",
            "leg2_index_lag": 2,
        },
        "usd_zcis": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "1+",
            "leg2_index_method": "daily",
            "leg2_index_lag": 3,
        },
        "gbp_zcs": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": True,
            "modifier": "mf",
            "calendar": "ldn",
            "payment_lag": 0,
            "currency": "gbp",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "sek_iirs": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "stk",
            "payment_lag": 0,
            "currency": "sek",
            "convention": "actacticma",
            "leg2_frequency": "q",
            "leg2_convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
            "index_method": "daily",
            "index_lag": 3,
        },
        "eur_sbs36": {
            "frequency": "q",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "act360",
            "leg2_frequency": "s",
            "spread_compound_method": "none_simple",
            "fixing_method": "ibor",
            "method_param": 2,
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "nok_sbs36": {
            "frequency": "q",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "osl",
            "payment_lag": 0,
            "currency": "nok",
            "convention": "act360",
            "leg2_frequency": "s",
            "spread_compound_method": "none_simple",
            "fixing_method": "ibor",
            "method_param": 2,
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "us_gb": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 1,
            "calc_mode": "us_gb",
        },
        "us_gb_tsy": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 1,
            "calc_mode": "us_gb_tsy",
        },
        "uk_gb": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "ldn",
            "payment_lag": 0,
            "currency": "gbp",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 7,
            "calc_mode": "uk_gb",
        },
        "de_gb": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 2,
            "ex_div": 1,
            "calc_mode": "de_gb",
        },
        "fr_gb": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 2,
            "ex_div": 1,
            "calc_mode": "fr_gb",
        },
        "nl_gb": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 2,
            "ex_div": 1,
            "calc_mode": "nl_gb",
        },
        "it_gb": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 2,
            "ex_div": 1,
            "calc_mode": "it_gb",
        },
        "se_gb": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "stk",
            "payment_lag": 0,
            "currency": "sek",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 5,
            "calc_mode": "se_gb",
        },
        "no_gb": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "osl",
            "payment_lag": 0,
            "currency": "nok",
            "convention": "actacticma_stub365f",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 1,
            "calc_mode": "no_gb",
        },
        "ca_gb": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "tro",
            "payment_lag": 0,
            "currency": "cad",
            "convention": "actacticma_stub365f",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 1,
            "calc_mode": "ca_gb",
        },
        "us_gbb": {
            "modifier": "none",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "act360",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 0,
            "calc_mode": "us_gbb",
        },
        "se_gbb": {
            "modifier": "none",
            "calendar": "stk",
            "payment_lag": 0,
            "currency": "sek",
            "convention": "act360",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 0,
            "calc_mode": "se_gbb",
        },
        "gbp_gbi": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "ldn",
            "payment_lag": 0,
            "currency": "gbp",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "index_method": "daily",
            "index_lag": 3,
            "settle": 1,
            "ex_div": 7,
            "calc_mode": "ukg",
        },
        "sek_fra3": {
            "termination": "3m",
            "frequency": "q",
            "eom": False,
            "modifier": "mf",
            "calendar": "stk",
            "payment_lag": 0,
            "currency": "sek",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "eur_fra3": {
            "termination": "3m",
            "frequency": "q",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "eur_fra6": {
            "termination": "6m",
            "frequency": "s",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "eur_fra1": {
            "termination": "1m",
            "frequency": "m",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "nok_fra3": {
            "termination": "3m",
            "frequency": "q",
            "eom": False,
            "modifier": "mf",
            "calendar": "osl",
            "payment_lag": 0,
            "currency": "nok",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "nok_fra6": {
            "termination": "6m",
            "frequency": "s",
            "eom": False,
            "modifier": "mf",
            "calendar": "osl",
            "payment_lag": 0,
            "currency": "nok",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "ibor",
            "leg2_method_param": 2,
        },
        "usd_frn5": {
            "frequency": "q",
            "eom": False,
            "modifier": "mf",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "act360",
            "spread_compound_method": "none_simple",
            "fixing_method": "rfr_observation_shift",
            "method_param": 5,
            "settle": 1,
            "ex_div": 1,
        },
        "usd_stir": {
            "frequency": "q",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "bp_value": 25.0,
            "nominal": 1000000.0,
        },
        "usd_stir1": {
            "frequency": "m",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay_avg",
            "bp_value": 41.67,
            "nominal": 5000000.0,
        },
        "eur_stir": {
            "frequency": "q",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "bp_value": 25.0,
            "nominal": 1000000.0,
        },
        "eur_stir1": {
            "frequency": "m",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay_avg",
            "bp_value": 25.0,
            "nominal": 3000000.0,
        },
        "eur_stir3": {
            "frequency": "q",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "bp_value": 25.0,
            "nominal": 1000000.0,
        },
        "gbp_stir": {
            "frequency": "q",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "ldn",
            "payment_lag": 0,
            "currency": "gbp",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "bp_value": 25.0,
            "nominal": 1000000.0,
        },
        "sofr": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "mf",
            "calendar": "nyc",
            "payment_lag": 2,
            "currency": "usd",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
        },
        "ust": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 1,
            "calc_mode": "us_gb",
        },
        "ustb": {
            "modifier": "none",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "act360",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 0,
            "calc_mode": "us_gbb",
        },
        "ukt": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "ldn",
            "payment_lag": 0,
            "currency": "gbp",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 7,
            "calc_mode": "uk_gb",
        },
        "gilt": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "ldn",
            "payment_lag": 0,
            "currency": "gbp",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 7,
            "calc_mode": "uk_gb",
        },
        "cadgb": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "tro",
            "payment_lag": 0,
            "currency": "cad",
            "convention": "actacticma_stub365f",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 1,
            "calc_mode": "ca_gb",
        },
        "sgb": {
            "frequency": "a",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "stk",
            "payment_lag": 0,
            "currency": "sek",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 5,
            "calc_mode": "se_gb",
        },
        "sgbb": {
            "modifier": "none",
            "calendar": "stk",
            "payment_lag": 0,
            "currency": "sek",
            "convention": "act360",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "settle": 1,
            "ex_div": 0,
            "calc_mode": "se_gbb",
        },
        "ukti": {
            "frequency": "s",
            "stub": "shortfront",
            "eom": False,
            "modifier": "none",
            "calendar": "ldn",
            "payment_lag": 0,
            "currency": "gbp",
            "convention": "actacticma",
            "initial_exchange": False,
            "final_exchange": True,
            "payment_lag_exchange": 0,
            "index_method": "daily",
            "index_lag": 3,
            "settle": 1,
            "ex_div": 7,
            "calc_mode": "ukg",
        },
        "sofr3mf": {
            "frequency": "q",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "bp_value": 25.0,
            "nominal": 1000000.0,
        },
        "sofr1mf": {
            "frequency": "m",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "nyc",
            "payment_lag": 0,
            "currency": "usd",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay_avg",
            "bp_value": 41.67,
            "nominal": 5000000.0,
        },
        "sonia3mf": {
            "frequency": "q",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "ldn",
            "payment_lag": 0,
            "currency": "gbp",
            "convention": "act365f",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "bp_value": 25.0,
            "nominal": 1000000.0,
        },
        "estr1mf": {
            "frequency": "m",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay_avg",
            "bp_value": 25.0,
            "nominal": 3000000.0,
        },
        "estr3mf": {
            "frequency": "q",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "bp_value": 25.0,
            "nominal": 1000000.0,
        },
        "euribor3mf": {
            "frequency": "q",
            "roll": "imm",
            "eom": False,
            "modifier": "mf",
            "calendar": "tgt",
            "payment_lag": 0,
            "currency": "eur",
            "convention": "act360",
            "leg2_spread_compound_method": "none_simple",
            "leg2_fixing_method": "rfr_payment_delay",
            "bp_value": 25.0,
            "nominal": 1000000.0,
        },
    }
